/* \file barrier
 * \brief A barrier for a group of threads
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.barrier;

import concurrent.sync;

/** \class Barrier
 * \brief  Barriers serve as synchronization points for groups of threads that
 * must occasionally wait for each other. 
 *
 * Barriers may support any of several methods that
 * accomplish this synchronization. This interface
 * merely expresses their minimal commonalities:
 *   - Every barrier is defined for a given number
 *     of <code>parties</code> -- the number of threads
 *     that must meet at the barrier point. (In all current
 *     implementations, this
 *     value is fixed upon construction of the Barrier.)
 *   - A barrier can become <code>broken</code> if
 *     one or more threads leave a barrier point prematurely,
 *     generally due to interruption or timeout. Corresponding
 *     synchronization methods in barriers fail, throwing
 *     BrokenBarrierException for other threads
 *     when barriers are in broken states.
 */
interface Barrier {

  /** 
   * Return the number of parties that must meet per barrier
   * point. The number of parties is always at least 1.
   */

  int parties();

  /**
   * Returns true if the barrier has been compromised
   * by threads leaving the barrier before a synchronization
   * point (normally due to interruption or timeout). 
   * Barrier methods in implementation classes throw
   * throw BrokenBarrierException upon detection of breakage.
   * Implementations may also support some means
   * to clear this status.
   */

  bool broken();
}

/** \class BrokenBarrierException
 * Thrown when a barrier has been compromised by
 * threads leaving the barrier before a synchronization point.
 */
class BrokenBarrierException : Exception {

  /** 
   * The index that barrier would have returned upon
   * normal return;
   */
  final int index;

  /**
   * Constructs a BrokenBarrierException with given index
   */
  this(int idx) {
    super("Broken barrier");
    index = idx;
  }

  /**
   * Constructs a BrokenBarrierException with the
   * specified index and detail message.
   */
  this(int idx, char[] message) {
    super(message);
    index = idx;
  }
}

/** \class TimeoutException
 * Thrown by synchronization classes that report
 * timeouts via exceptions. The exception is treated
 * as a form (subclass) of WaitException. This both
 * simplifies handling, and conceptually reflects the fact that
 * timed-out operations are artificially interrupted by timers.
 */
class TimeoutException : WaitException {
  /** 
   * The approximate time that the operation lasted before 
   * this timeout exception was thrown.
   */
  public final long duration;

  /**
   * Constructs a TimeoutException with given duration value.
   */
  this(long time) {
    super("Timeout exception");
    duration = time;
  }

  /**
   * Constructs a TimeoutException with the
   * specified duration value and detail message.
   */
  this(long time, char[] message) {
    super(message);
    duration = time;
  }
}
