/* \file countdown
 * \brief A latch that fires after a specified count
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
*/

module concurrent.countdown;

import concurrent.sync;

private import concurrent.waitnotify;
private import std.c.time;

private import std.thread; // for unittest

/** \class CountDown
 * \brief A CountDown can serve as a simple one-shot barrier. 
 *
 * A Countdown is initialized
 * with a given count value. Each release decrements the count.
 * All acquires block until the count reaches zero. Upon reaching
 * zero all current acquires are unblocked and all 
 * subsequent acquires pass without blocking. This is a one-shot
 * phenomenon -- the count cannot be reset. 
 * If you need a version that resets the count, consider
 * using a Barrier.
 * <p>
 * <b>Sample usage.</b> Here are a set of classes in which
 * a group of worker threads use a countdown to
 * notify a driver when all threads are complete.
 * \code
 * class Worker { 
 *   private final CountDown done;
 *   this(CountDown d) { done = d; }
 *   int run() {
 *     doWork();
 *    done.release();
 *   }
 * }
 * 
 * void test() {
 *   CountDown done = new CountDown(N);
 *   for (int i = 0; i < N; ++i) {
 *     Worker w = new Worker(done);
 *     new Thread(&w.run).start();
 *   }
 *   for (int k=0;k<10000; k++){}
 *   done.acquire(); // wait for all to finish
 * }
 * \endcode
 *
 */

class CountDown : Sync {
  protected final int initialCount_;
  protected int count_;
  protected SimpleWaitNotifyAll wnlock;

  /** Create a new CountDown with given count value */
  this(int count) { 
    count_ = initialCount_ = count; 
    wnlock.init();
  }

  /** Destroy a CountDown and release system resources */
  ~this() { 
    wnlock.destroy();
  }

  /*
    This could use double-check, but doesn't out of concern
    for surprising effects on user programs stemming
    from lack of memory barriers with lack of synch.
  */
  synchronized void acquire() {
    while (count_ > 0) 
      wnlock.wait(this);
  }


  synchronized bool attempt(long msecs) {
    if (count_ <= 0) 
      return true;
    else if (msecs <= 0) 
      return false;
    else {
      long waitTime = msecs;
      long start = clock();
      for (;;) {
	wnlock.wait(this,waitTime);
	if (count_ <= 0) 
	  return true;
	else {
	  waitTime = msecs - (clock() - start);
	  if (waitTime <= 0) 
	    return false;
	}
      }
    }
  }

  /**
   * Decrement the count.
   * After the initialCount'th release, all current and future
   * acquires will pass
   */
  synchronized void release() {
    if (--count_ == 0) 
      wnlock.notifyAll();
  }

  /** Return the initial count value */
  int initialCount() { return initialCount_; }

  /** 
   * Return the current count value.
   * This is just a snapshot value, that may change immediately
   * after returning.
   */
  synchronized int currentCount() { return count_; }

  unittest {
    class Worker { 
      private final CountDown done;
      this(CountDown d) { done = d; }
      int run() {
	printf("counting down...\n");
	for (int k=0;k<10000; k++){}
	done.release();
	return 0;
      }
    }
    int N = 5;
    printf("starting countdown unittest\n");
    CountDown done = new CountDown(N);
    for (int i = 0; i < N; ++i) {
      Worker w = new Worker(done);
      Thread t = new Thread(&w.run);
      t.start();
    }
    for (int k=0;k<10000; k++){}
    done.acquire(); // wait for all to finish
    printf("finished countdown unittest\n");
  }
}

