/** \file fairsemaphore.d
 * \brief Counting semaphore with enforced fairness
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.fairsemaphore;

import concurrent.semaphore;

private import concurrent.waitnotify;
private import std.c.time;

private import std.thread; // for unittest


/** \class FairSemaphore
 * \brief An implementation of counting Semaphores that
 *  enforces enough fairness for applications that
 *  need to avoid indefinite overtaking without
 *  necessarily requiring FIFO ordered access.
 *
 *  Empirically, very little is paid for this property
 *  unless there is a lot of contention among threads
 *  or very unfair native thread scheduling.
 *  The acquire method waits even if there are permits
 *  available but have not yet been claimed by threads that have
 *  been notified but not yet resumed. This makes the semaphore
 *  almost as fair as the underlying primitives allow. 
 *  So, if synch lock entry and notify are both fair
 *  so is the semaphore -- almost:  Rewaits stemming
 *  from timeouts in attempt, along with potentials for
 *  interrupted threads to be notified can compromise fairness,
 *  possibly allowing later-arriving threads to pass before
 *  later arriving ones. However, in no case can a newly
 *  entering thread obtain a permit if there are still others waiting.
 *  Also, signalling order need not coincide with
 *  resumption order. Later-arriving threads might get permits
 *  and continue before other resumable threads are actually resumed.
 *  However, all of these potential fairness breaches are
 *  very rare in practice unless the underlying native threads
 *  performs strictly LIFO notifications in which case you need to use
 *  a FIFOSemaphore to maintain a reasonable approximation
 *  of fairness.
 * 
*/

final class FairSemaphore : Semaphore  {

  /** 
   * Create a Semaphore with the given initial number of permits.
  */
  this(long initial) {  super(initial); }

  protected long waits_ = 0;   ///< Number of waiting threads

  synchronized void acquire() {
    /*
      Only take if there are more permits than threads waiting
      for permits. This prevents infinite overtaking.
    */ 
    if (permits_ > waits_) { 
      --permits_;
      return;
    }
    else { 
      ++waits_;
      try { 
	for (;;) {
	  wnlock.wait(this); 
	  if (permits_ > 0) {
	    --waits_;
	    --permits_;
	    return;
	  }
	}
      }
      catch(WaitException ex) { 
	--waits_;
	wnlock.notify();
	throw ex;
      }
    }
  }

  synchronized bool attempt(long msecs) {
    if (permits_ > waits_) { 
      --permits_;
      return true;
    }
    else if (msecs <= 0)   
      return false;
    else {
      ++waits_;
        
      long startTime = clock();
      long waitTime = msecs;
        
      try {
	for (;;) {
	  wnlock.wait(this,waitTime);
	  if (permits_ > 0) {
	    --waits_;
	    --permits_;
	    return true;
	  }
	  else { // got a time-out or false-alarm notify
	    waitTime = msecs - (clock() - startTime);
	    if (waitTime <= 0) {
	      --waits_;
	      return false;
	    }
	  }
	}
      }
      catch(WaitException ex) { 
	--waits_;
	wnlock.notify();
	throw ex;
      }
    }
  }

  synchronized void release() {
    ++permits_;
    wnlock.notify();
  }

  /** Release N permits */
  synchronized void release(long n) {
    permits_ += n;
    for (long i = 0; i < n; ++i) wnlock.notify();
  }

  unittest {
    FairSemaphore sem = new FairSemaphore(2);
    int done = 0;
    Thread[] t = new Thread[6];
    int f() {
      int n;
      Thread tt = Thread.getThis();
      for (n=0; n < t.length; n++) {
	if (tt is t[n])
	  break;
      }
      printf(" thread %d started\n",n);
      sem.acquire();
      printf(" thread %d aquired\n",n);
      return 0;
    }
    int f2() {
      int n;
      Thread tt = Thread.getThis();
      for (n=0; n < t.length; n++) {
	if (tt is t[n])
	  break;
      }
      printf(" thread %d releasing\n",n);
      sem.release();
      return 0;
    }
    int n;
    for (n=0; n<t.length/2; n++) {
      t[n] = new Thread(&f);
    }
    for (; n<t.length; n++) {
      t[n] = new Thread(&f2);
    }
    printf("starting fair semaphore unittest\n");
    for (n=0; n<t.length/2; n++) {
      t[n].start();
    }
    Thread.yield();
    for (; n<t.length; n++) {
      t[n].start();
      Thread.yield();
    }
    printf("finished fair semaphore unittest\n");
    delete sem;
    t[] = null;
  }
}

