/** \file fifosemaphore.d
 * \brief First-in/First-out semaphore.
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.fifosemaphore;

import concurrent.queuedsemaphore;

private import concurrent.waitnotify;
private import std.c.time;

private import std.thread; // for unittest

/** \class FIFOSemaphore
 * \brief A First-in/First-out implementation of a Semaphore.
 *
 * Waiting requests will be satisified in
 * the order that the processing of those requests got to a certain point.
 * If this sounds vague it is meant to be. FIFO implies a
 * logical timestamping at some point in the processing of the
 * request. To simplify things we don't actually timestamp but
 * simply store things in a FIFO queue. Thus the order in which
 * requests enter the queue will be the order in which they come
 * out.  This order need not have any relationship to the order in
 * which requests were made, nor the order in which requests
 * actually return to the caller.
 */

class FIFOSemaphore : QueuedSemaphore {
  
  /** 
   * Create a Semaphore with the given initial number of permits.
   * Using a seed of one makes the semaphore act as a mutual exclusion lock.
   * Negative seeds are also allowed, in which case no acquires will proceed
   * until the number of releases has pushed the number of permits past 0.
  */

  this(long initialPermits) { 
    super(new FIFOWaitQueue(), initialPermits);
  }

  /** 
   * Simple linked list queue used in FIFOSemaphore.
   * Methods are not synchronized; they depend on synch of callers
  */

  protected class FIFOWaitQueue : WaitQueue {
    protected WaitNode head_ = null;
    protected WaitNode tail_ = null;

    protected void insert(WaitNode w) {
      if (tail_ is null) 
        head_ = tail_ = w;
      else {
        tail_.next = w;
        tail_ = w;
      }
    }

    protected WaitNode extract() { 
      if (head_ is null) 
        return null;
      else {
        WaitNode w = head_;
        head_ = w.next;
        if (head_ is null) tail_ = null;
        w.next = null;  
        return w;
      }
    }
  }
  unittest {
    FIFOSemaphore sem = new FIFOSemaphore(3);
    int done = 0;
    Thread[] t = new Thread[10];
    int f() {
      int n;
      Thread tt = Thread.getThis();
      for (n=0; n < t.length; n++) {
	if (tt is t[n])
	  break;
      }
      sem.acquire();
      return 0;
    }
    int f2() {
      int n;
      Thread tt = Thread.getThis();
      for (n=0; n < t.length; n++) {
	if (tt is t[n])
	  break;
      }
      sem.release();
      return 0;
    }
    int n;
    for (n=0; n<t.length/2; n++) {
      t[n] = new Thread(&f);
    }
    for (; n<t.length; n++) {
      t[n] = new Thread(&f2);
    }
    printf("starting fifosemaphore unittest\n");
    for (n=0; n<t.length/2; n++) {
      t[n].start();
    }
    Thread.yield();
    for (; n<t.length; n++) {
      t[n].start();
      Thread.yield();
    }
    printf("finished fifosemaphore unittest\n");
    delete sem;
    t[] = null;
  }

}
