/* \file latch
 * \brief A condition that is acquirable forever after the first time.
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.latch;

import concurrent.sync;

private import concurrent.waitnotify;
private import std.c.time;

private import std.thread; // for unittest

/** \class Latch
 * \brief A latch is a boolean condition that is set at most once, ever.
 * Once a single release is issued, all acquires will pass.
 *
 * <b>Sample usage.</b> Here are a set of classes that use
 * a latch as a start signal for a group of worker threads that
 * are created and started beforehand, and then later enabled.
 * \code
 * class Worker implements Runnable {
 *   private final Latch startSignal;
 *   Worker(Latch l) { startSignal = l; }
 *    int run() {
 *      startSignal.acquire();
 *      for (int =0; i<10000; i++){} // do something
 *      printf(" thread done\n");
 *      return 0;
 *   }
 * }
 *
 * class Driver { // ...
 *   void main() {
 *     Latch go = new Latch();
 *     for (int i = 0; i < N; ++i) {// make threads
 *       Worker w = new Worker(go);
 *       new Thread(&w.run).start();
 *     }
 *     printf("pausing\n");
 *     for (int i=0; i<10000; i++) {} // pause
 *     printf("let go\n");
 *     go.release();              // let all threads proceed
 *   } 
 * }
 * \endcode
 */  

class Latch : Sync {
  protected bool latched_ = false;
  protected SimpleWaitNotifyAll wnlock;

  this() { wnlock.init(); }
  ~this() { wnlock.destroy(); }

  /*
    This could use double-check, but doesn't.
    If the latch is being used as an indicator of
    the presence or state of an object, the user would
    not necessarily get the memory barrier that comes with synch
    that would be needed to correctly use that object. This
    would lead to errors that users would be very hard to track down. So, to
    be conservative, we always use synch.
  */

  synchronized void acquire() {
    while (!latched_) 
      wnlock.wait(this); 
  }

  synchronized bool attempt(long msecs) {
    if (latched_) 
      return true;
    else if (msecs <= 0) 
      return false;
    else {
      long waitTime = msecs;
      long start = clock();
      for (;;) {
	wnlock.wait(this,waitTime);
	if (latched_) 
	  return true;
	else {
	  waitTime = msecs - (clock() - start);
	  if (waitTime <= 0) 
	    return false;
	}
      }
    }
  }

  /** Enable all current and future acquires to pass */
  synchronized void release() {
    latched_ = true;
    wnlock.notifyAll();
  }

  unittest {
    class Worker {
      private final Latch startSignal;
      this(Latch l) { startSignal = l; }
      int run() {
	startSignal.acquire();
	printf(" thread done\n");
	return 0;
      }
    }
 
    printf("starting latch unittest\n");
    Latch go = new Latch();
    for (int i = 0; i < 10; ++i) {// make threads
      Worker w = new Worker(go);
      Thread t = new Thread(&w.run);
      t.start();
    }
    printf("threads started\n");
    for (int i=0; i<10000; i++) {} // pause
    printf("release latch\n");
    go.release();              // let all threads proceed
    for (int i=0; i<100; i++) {
      Thread.yield();
    }
    printf("finished latch unittest\n");
  }
}

