/* \file linkedqueue
 * \brief Basic Channel implementation
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.

  History:
  Date       Who                What
  11Jun1998  dl               Create public version
  25aug1998  dl               added peek
  10dec1998  dl               added isEmpty
  10oct1999  dl               lock on node object to ensure visibility
	07May2004  Mike Swieton     Translated to D
*/

module concurrent.linkedqueue;

import std.c.time;

import concurrent.channel;
import concurrent.linkednode;
import concurrent.waitnotify;

/**
 * A linked list based channel implementation.
 * The algorithm avoids contention between puts
 * and takes when the queue is not empty. 
 * Normally a put and a take can proceed simultaneously. 
 * (Although it does not allow multiple concurrent puts or takes.)
 * This class tends to perform more efficently than
 * other Channel implementations in producer/consumer
 * applications.
 **/

class LinkedQueue(T) : Channel!(T) {

	protected alias LinkedNode!(T) node_type;

  /** 
   * Dummy header node of list. The first actual node, if it exists, is always 
   * at head_.next. After each take, the old first node becomes the head.
   **/
  protected node_type head_;         

  /**
   * Helper monitor for managing access to last node.
   **/
  protected final Object putLock_;

	/**
	 * Wait lock, used because D's Object doesn't have wait()/notify semantics.
	 **/
	protected SimpleWaitNotify waitLock_;

  /** 
   * The last node of list. Put() appends to list, so modifies last_
   **/
  protected node_type last_;         

  /**
   * The number of threads waiting for a take.
   * Notifications are provided in put only if greater than zero.
   * The bookkeeping is worth it here since in reasonably balanced
   * usages, the notifications will hardly ever be necessary, so
   * the call overhead to notify can be eliminated.
   **/
  protected int waitingForTake_ = 0;  

  public this() {
 		putLock_ = new Object(); 
    head_ = new node_type; 
    last_ = head_;

		waitLock_.init();
  }

  /** Main mechanics for put/offer **/
  protected void insert(T x)
	in {
		assert(x !== null);
	} body { 
    synchronized(putLock_) {
      node_type p = new node_type(x);
      synchronized(last_) {
        last_.next = p;
        last_ = p;
      }
      if (waitingForTake_ > 0)
        waitLock_.notify();
    }
  }

  /** Main mechanics for take/poll **/
  protected synchronized T extract() {
    synchronized(head_) {
      T x = null;
      node_type first = head_.next;
      if (first !== null) {
        x = first.value;
        first.value = null;
        head_ = first; 
      }
      return x;
    }
  }


  public void put(T x)
	in {
		assert(x !== null); 
	} body {
    insert(x); 
  }

  public bool offer(T x, long msecs)
	in {
		assert(x !== null);
		assert(msecs >= 0);
	} body { 
    insert(x); 
    return true;
  }

	public T take() {
		// try to extract. If fail, then enter wait-based retry loop
		T x = extract();
		if (x !== null)
			return x;
		else { 
			synchronized(putLock_) {
				++waitingForTake_;
				for (;;) {
					x = extract();
					if (x !== null) {
						--waitingForTake_;
						return x;
					}
					else {
						try {
							waitLock_.wait(putLock_); 
						} catch (WaitException ex)
						{
							waitLock_.notify();
							throw ex;
						}
					}
				}
			}
		}
	}

  public T peek() {
    synchronized(head_) {
      node_type first = head_.next;
      if (first !== null) 
        return first.value;
      else 
        return null;
    }
  }    


  public bool isEmpty() {
    synchronized(head_) {
			return head_.next is null;
		}
	}    

	public T poll(long msecs)
	in {
		assert(msecs >= 0);
	} body {
		T x = extract();
		if (x !== null) 
			return x;
		else {
			synchronized(putLock_) {
				long waitTime = msecs;
				long start = (msecs <= 0)? 0 : clock();
				++waitingForTake_;
				for (;;) {
					x = extract();
					if (x !== null || waitTime <= 0) {
						--waitingForTake_;
						return x;
					}
					else {
						try
						{
							waitLock_.wait(putLock_, waitTime); 
							waitTime = msecs - (clock() - start);
						} catch (WaitException ex) {
							waitLock_.notify();
							throw ex;
						}
					}
				}
			}
		}
	}
}
