/* \file lockedexecutor
 * \brief Runs commands in a mutex lock
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.

  History:
  Date       Who                What
	01May2004  Mike Swieton     Translated to D
  21Jun1998  dl               Create public version
*/

module concurrent.lockedexecutor;

import concurrent.executor;
import concurrent.sync;

/**
 * An implementation of Executor that 
 * invokes the run method of the supplied command within
 * a synchronization lock and then returns.
 * 
 **/
class LockedExecutor : Executor {
  
  /** The mutex **/
	protected:
		Sync mutex_;

	public:

  /** 
   * Create a new LockedExecutor that relies on the given mutual
   * exclusion lock. 
   * @param mutex Any mutual exclusion lock.
   * Standard usage is to supply an instance of <code>Mutex</code>,
   * but, for example, a Semaphore initialized to 1 also works.
   * On the other hand, many other Sync implementations would not
   * work here, so some care is required to supply a sensible 
   * synchronization object.
   **/
  this(Sync mutex) {
		mutex_ = mutex;
  }

  /** 
   * Execute the given command directly in the current thread,
   * within the supplied lock.
   **/
  void execute(int delegate() command) {
    mutex_.acquire();
    try {
      command();
    }
    finally {
      mutex_.release();
    }
  }
}
