/* \file mutex
 * \brief Mutual exlusive lock
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.mutex;

import concurrent.sync;

private import concurrent.waitnotify;
private import std.c.time;

private import std.thread; // for unittest

/** \class Mutex
 * \brief A simple non-reentrant mutual exclusion lock.
 * 
 * The lock is free upon construction. Each acquire gets the
 * lock, and each release frees it. Releasing a lock that
 * is already free has no effect. 
 * 
 * This implementation makes no attempt to provide any fairness
 * or ordering guarantees. If you need them, consider using one of
 * the Semaphore implementations as a locking mechanism.
 * 
 * <b>Sample usage</b><br>
 * 
 * Mutex can be useful in constructions that cannot be
 * expressed using synchronized blocks because the
 * acquire/release pairs do not occur in the same method or
 * code block. For example, you can use them for hand-over-hand
 * locking across the nodes of a linked list. This allows
 * extremely fine-grained locking,  and so increases 
 * potential concurrency, at the cost of additional complexity and
 * overhead that would normally make this worthwhile only in cases of
 * extreme contention.
 * \code
 * class Node { 
 *   Object item; 
 *   Node next; 
 *   Mutex lock = new Mutex(); // each node keeps its own lock
 *
 *   Node(Object x, Node n) { item = x; next = n; }
 * }
 *
 * class List {
 *    protected Node head; // pointer to first node of list
 *
 *    // Use plain synchronization to protect head field.
 *    //  (We could instead use a Mutex here too but there is no
 *    //  reason to do so.)
 *    protected synchronized Node getHead() { 
 *      return head; 
 *    }
 *
 *    bool search(Object x) {
 *      Node p = getHead();
 *      if (p is null) return false;
 *
 *      //  (This could be made more compact, but for clarity of illustration,
 *      //  all of the cases that can arise are handled separately.)
 *
 *      p.lock.acquire();              // Prime loop by acquiring first lock.
 *                                     //    (If the acquire fails due to
 *                                     //    interrupt, the method will throw
 *                                     //    WaitException now,
 *                                     //    so there is no need for any
 *                                     //    further cleanup.)
 *      for (;;) {
 *        if (x == p.item) {
 *          p.lock.release();          // release current before return
 *          return true;
 *        }
 *        else {
 *          Node nextp = p.next;
 *          if (nextp is null) {
 *            p.lock.release();       // release final lock that was held
 *            return false;
 *          }
 *          else {
 *            try {
 *              nextp.lock.acquire(); // get next lock before releasing current
 *            }
 *            catch (WaitException ex) {
 *              p.lock.release();    // also release current if acquire fails
 *              throw ex;
 *            }
 *            p.lock.release();      // release old lock now that new one held
 *            p = nextp;
 *          }
 *        }
 *      }
 *    }
 *
 *    void synchronized add(Object x) { // simple prepend
 *      // The use of `synchronized'  here protects only head field.
 *      // The method does not need to wait out other traversers 
 *      // who have already made it past head.
 *      head = new Node(x, head);
 *    }
 *
 *    // ...  other similar traversal and update methods ...
 * }
 * \endcode
 *
 */

class Mutex : Sync  {

  protected bool inuse_ = false; ///< The lock status

  protected SimpleWaitNotify wnlock;

  this() {
    wnlock.init();
  }

  ~this() {
    wnlock.destroy();
  }

  synchronized void acquire() {
    try {
      while (inuse_) wnlock.wait(this);
      inuse_ = true;
    }
    catch (WaitException ex) {
      wnlock.notify();
      throw ex;
    }
  }

  synchronized void release()  {
    inuse_ = false;
    wnlock.notify(); 
  }


  synchronized bool attempt(long msecs) {
    if (!inuse_) {
      inuse_ = true;
      return true;
    }
    else if (msecs <= 0)
      return false;
    else {
      long waitTime = msecs;
      long start = clock();
      try {
	for (;;) {
	  wnlock.wait(this,waitTime);
	  if (!inuse_) {
	    inuse_ = true;
	    return true;
	  }
	  else {
	    waitTime = msecs - (clock() - start);
	    if (waitTime <= 0) 
	      return false;
	  }
	}
      }
      catch (WaitException ex) {
	wnlock.notify();
	throw ex;
      }
    }
  }

  unittest {
    class Test {
      Mutex lock;
      int acquired;
      Thread[] t;
      int f() {
	int n;
	Thread tt = Thread.getThis();
	for (n=0; n < t.length; n++) {
	  if (tt is t[n])
	    break;
	}
	printf(" thread %d started\n",n);
	lock.acquire();
	printf(" thread %d aquired\n",n);
	lock.release();
	acquired++;
	printf(" thread %d released\n",n);
	return 0;
      }
      void run() {
	lock = new Mutex();
	acquired = 0;
	t = new Thread[3];
	int n;
	for (n=0; n<t.length; n++) {
	  t[n] = new Thread(&this.f);
	}
	printf("starting mutex unittest\n");
	for (n=0; n<t.length; n++) {
	  t[n].start();
	}
	while (acquired != n)
	  Thread.yield();
	printf("finished mutex unittest\n");
      }
    }
    Test t = new Test();
    t.run();
  }
}
