/* \file puttable
 * \brief Interface for input to inter-thread queues
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.

  History:
  Date       Who                What
  11Jun1998  dl               Create public version
	07May2004  Mike Swieton     Translated to D
*/

module concurrent.puttable;

/** 
 * This interface exists to enable stricter type checking for channels. A
 * method argument or instance variable in a producer object can be declared as
 * only a Puttable rather than a Channel, in which case a compiler will
 * disallow take operations.
 * <p>
 * Full method descriptions appear in the Channel interface.
 * @see Channel
 * @see Takable
**/

interface Puttable(T) {

  /** 
   * Place item in the channel, possibly waiting indefinitely until
   * it can be accepted. Channels implementing the BoundedChannel
   * subinterface are generally guaranteed to block on puts upon
   * reaching capacity, but other implementations may or may not block.
   * @param item the element to be inserted. Should be non-null.
  **/
  public void put(T item);


  /** 
   * Place item in channel only if it can be accepted within
   * msecs milliseconds. The time bound is interpreted in
   * a coarse-grained, best-effort fashion. 
   * @param item the element to be inserted. Should be non-null.
   * @param msecs the number of milliseconds to wait. If less than
   * or equal to zero, the method does not perform any timed waits,
	 * but might still require access to a synchronization lock, which can impose
	 * unbounded delay if there is a lot of contention for the channel.
   * @return true if accepted, else false
  **/
  public bool offer(T item, long msecs);
}
