/** \file queuedsemaphore.d
 * \brief Abstract semaphore with queued wait nodes.
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.queuedsemaphore;

import concurrent.semaphore;

private import concurrent.waitnotify;
private import std.c.time;

private import std.thread; // for unittest

/** \class QueuedSemaphore
 * Abstract base class for semaphores relying on queued wait nodes.
 */
class QueuedSemaphore : Semaphore {
  
  protected final WaitQueue wq_;

  this(WaitQueue q, long initialPermits) { 
    super(initialPermits);  
    wq_ = q;
  }

  void acquire() {
    if (precheck()) return;
    WaitNode w = new WaitNode();
    w.doWait(this);
  }

  bool attempt(long msecs) {
    if (precheck()) return true;
    if (msecs <= 0) return false;

    WaitNode w = new WaitNode();
    return w.doTimedWait(this, msecs);
  }

  protected synchronized bool precheck() {
    bool pass = (permits_ > 0);
    if (pass) --permits_;
    return pass;
  }

  protected synchronized bool recheck(WaitNode w) {
    bool pass = (permits_ > 0);
    if (pass) --permits_;
    else      wq_.insert(w);
    return pass;
  }


  protected synchronized WaitNode getSignallee() {
    WaitNode w = wq_.extract();
    if (w is null) ++permits_; // if none, inc permits for new arrivals
    return w;
  }

  void release() {
    for (;;) {
      WaitNode w = getSignallee();
      if (w is null) return;  // no one to signal
      if (w.signal()) return; // notify if still waiting, else skip
    }
  }

  /** Release N permits */
  void release(long n) 
  in {
    assert(n >= 0);
  }
  body {
    for (long i = 0; i < n; ++i) release();
  }

  /** 
   * Base class for internal queue classes for semaphores, etc.
   * Relies on subclasses to actually implement queue mechanics
   */
  interface WaitQueue {
    void insert(WaitNode w);// assumed not to block
    WaitNode extract();     // should return null if empty
  }

}

class WaitNode {
  bool waiting = true;
  WaitNode next = null;
  protected SimpleWaitNotify wnlock;
  this() {
    wnlock.init(); 
  }
  ~this() {
    wnlock.destroy(); 
  }
  protected synchronized bool signal() {
    bool signalled = waiting;
    if (signalled) {
      waiting = false;
      wnlock.notify();
    }
    return signalled;
  }

  protected synchronized bool doTimedWait(QueuedSemaphore sem, 
					  long msecs)  {
    if (sem.recheck(this) || !waiting) 
      return true;
    else if (msecs <= 0) {
      waiting = false;
      return false;
    }
    else { 
      long waitTime = msecs;
      long start = clock();

      try {
	for (;;) {
	  wnlock.wait(this,waitTime);  
	  if (!waiting)   // definitely signalled
	    return true;
	  else { 
	    waitTime = msecs - (clock() - start);
	    if (waitTime <= 0) { //  timed out
	      waiting = false;
	      return false;
	    }
	  }
	}
      }
      catch(WaitException ex) {
	if (waiting) { // no notification
	  waiting = false; // invalidate for the signaller
	  throw ex;
	}
      }
    }
  }

  protected synchronized void doWait(QueuedSemaphore sem) {
    if (!sem.recheck(this)) {
      try {
	while (waiting) wnlock.wait(this);  
      }
      catch(WaitException ex) {
	if (waiting) { // no notification
	  waiting = false; // invalidate for the signaller
	  throw ex;
	}
      }
    }
  }
}
