/* \file reentrantlock
 * \brief A re-entrant lock like the builtin synchronization lock
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.reentrantlock;

import concurrent.sync;

private import concurrent.waitnotify;
private import std.c.time;

private import std.thread; // for unittest

/** \class ReentrantLock
 * \brief A lock with the same semantics as builtin synchronized locks: 
 *
 * Once a thread has a lock, it can re-obtain it any number of times
 * without blocking.  The lock is made available to other threads when
 * as many releases as acquires have occurred.
 */
class ReentrantLock : Sync  {

  protected Thread owner_ = null;
  protected long holds_ = 0;
  protected SimpleWaitNotify wnlock;

  this() { wnlock.init(); }

  ~this() { wnlock.destroy(); }

  void acquire() {
    Thread caller = Thread.getThis();
    synchronized(this) {
      if (caller == owner_) 
        ++holds_;
      else {
        try {  
          while (owner_ != null) wnlock.wait(this); 
          owner_ = caller;
          holds_ = 1;
        }
        catch (WaitException ex) {
          wnlock.notify();
          throw ex;
        }
      }
    }
  }  

  bool attempt(long msecs) {
    Thread caller = Thread.getThis();
    synchronized(this) {
      if (caller == owner_) {
        ++holds_;
        return true;
      }
      else if (owner_ == null) {
        owner_ = caller;
        holds_ = 1;
        return true;
      }
      else if (msecs <= 0)
        return false;
      else {
        long waitTime = msecs;
        long start = clock();
        try {
          for (;;) {
            wnlock.wait(this,waitTime); 
            if (caller == owner_) {
              ++holds_;
              return true;
            }
            else if (owner_ == null) {
              owner_ = caller;
              holds_ = 1;
              return true;
            }
            else {
              waitTime = msecs - (clock() - start);
              if (waitTime <= 0) 
                return false;
            }
          }
        }
        catch (WaitException ex) {
          wnlock.notify();
          throw ex;
        }
      }
    }
  }  

  /**
   * Release the lock.
   */
  synchronized void release()  
  in {
    assert( Thread.getThis() is owner_);
  }
  body {
    if (--holds_ == 0) {
      owner_ = null;
      wnlock.notify(); 
    }
  }

  /** 
   * Release the lock N times. <code>release(n)</code> is
   * equivalent in effect to:
   * \code
   *   for (int i = 0; i < n; ++i) release();
   * \endcode
   */
  synchronized void release(long n) 
  in {
    assert( Thread.getThis() is owner_);
  }
  body {
    holds_ -= n;
    if (holds_ == 0) {
      owner_ = null;
      wnlock.notify(); 
    }
  }

  /**
   * Return the number of unreleased acquires performed
   * by the current thread.
   * Returns zero if current thread does not hold lock.
   */
  synchronized long holds() {
    if (Thread.getThis() != owner_) return 0;
    return holds_;
  }
}

