/** \file rendezvous.d
 * \brief A barrier with abritrary threads and information exchange.
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.rendezvous;

import concurrent.barrier;

private import concurrent.fairsemaphore;
private import concurrent.waitnotify;
private import std.thread;
private import std.c.time;

/** \class Rendezvous
 * \brief A barrier with  information exchange.
 *
 * A rendezvous is a barrier that:
 *   - Unlike a CyclicBarrier, is not restricted to use 
 *     with fixed-sized groups of threads.
 *     Any number of threads can attempt to enter a rendezvous,
 *     but only the predetermined number of parties enter
 *     and later become released from the rendezvous at any give time.
 *   - Enables each participating thread to exchange information
 *     with others at the rendezvous point. Each entering thread
 *     presents some object on entry to the rendezvous, and
 *     returns some object on release. The object returned is
 *     the result of a RendezvousFunction that is run once per
 *     rendezvous, (it is run by the last-entering thread). By
 *     default, the function applied is a rotation, so each
 *     thread returns the object given by the next (modulo parties)
 *     entering thread. This default function faciliates simple
 *     application of a common use of rendezvous, as exchangers.
 *
 * Rendezvous use an all-or-none breakage model
 * for failed synchronization attempts: If threads
 * leave a rendezvous point prematurely because of timeout
 * or interruption, others will also leave abnormally
 * (via BrokenBarrierException), until
 * the rendezvous is <code>restart</code>ed. This is usually
 * the simplest and best strategy for sharing knowledge
 * about failures among cooperating threads in the most
 * common usages contexts of Rendezvous.
 *
 * While any positive number (including 1) of parties can
 * be handled, the most common case is to have two parties.
 *
 * <b>Sample Usage</b><p>
 * Here are the highlights of a class that uses a Rendezvous to
 * swap buffers between threads so that the thread filling the
 * buffer  gets a freshly
 * emptied one when it needs it, handing off the filled one to
 * the thread emptying the buffer.
 * \code
 * class FillAndEmpty {
 *   Rendezvous exchanger = new Rendezvous(2);
 *   Buffer initialEmptyBuffer = ... a made-up type
 *   Buffer initialFullBuffer = ...
 *
 *   class FillingLoop implements Runnable {
 *     public void run() {
 *       Buffer currentBuffer = initialEmptyBuffer;
 *       try {
 *         while (currentBuffer != null) {
 *           addToBuffer(currentBuffer);
 *           if (currentBuffer.full()) 
 *             currentBuffer = (Buffer)(exchanger.rendezvous(currentBuffer));
 *         }
 *       }
 *       catch (BrokenBarrierException ex) {
 *         return;
 *       }
 *       catch (WaitException ex) {
 *         Thread.currentThread().interrupt();
 *       }
 *     }
 *   }
 *
 *   class EmptyingLoop implements Runnable {
 *     public void run() {
 *       Buffer currentBuffer = initialFullBuffer;
 *       try {
 *         while (currentBuffer != null) {
 *           takeFromBuffer(currentBuffer);
 *           if (currentBuffer.empty()) 
 *             currentBuffer = (Buffer)(exchanger.rendezvous(currentBuffer));
 *         }
 *       }
 *       catch (BrokenBarrierException ex) {
 *         return;
 *       }
 *       catch (WaitException ex) {
 *         Thread.currentThread().interrupt();
 *       }
 *     }
 *   }
 *
 *   void start() {
 *     new Thread(new FillingLoop()).start();
 *     new Thread(new EmptyingLoop()).start();
 *   }
 * }
 */

class Rendezvous : Barrier {

  /**
   * Perform some function on the objects presented at
   * a rendezvous. The objects array holds all presented
   * items; one per thread. Its length is the number of parties. 
   * The array is ordered by arrival into the rendezvous.
   * So, the last element (at objects[objects.length-1])
   * is guaranteed to have been presented by the thread performing
   * this function. No identifying information is
   * otherwise kept about which thread presented which item.
   * If you need to 
   * trace origins, you will need to use an item type for rendezvous
   * that includes identifying information. After return of this
   * function, other threads are released, and each returns with
   * the item with the same index as the one it presented.
   */
  alias void function(Object[] objects) RendezvousFunction;

  /**
   * The default rendezvous function. Rotates the array
   * so that each thread returns an item presented by some
   * other thread (or itself, if parties is 1).
   */
  static void rendezvousFunction(Object[] objects) {
    int lastIdx = objects.length - 1;
    Object first = objects[0];
    for (int i = 0; i < lastIdx; ++i) objects[i] = objects[i+1];
    objects[lastIdx] = first;
  }

  protected final int parties_;

  protected bool broken_ = false;

  /** 
   * Number of threads that have entered rendezvous
   */
  protected int entries_ = 0;

  /** 
   * Number of threads that are permitted to depart rendezvous 
   */
  protected long departures_ = 0;

  /** 
   * Incoming threads pile up on entry until last set done.
   */
  protected final Semaphore entryGate_;

  /**
   * Temporary holder for items in exchange
   */
  protected final Object[] slots_;

  /**
   * The function to run at rendezvous point
   */

  protected RendezvousFunction rendezvousFunction_;

  protected SimpleWaitNotifyAll wnlock;

  /** 
   * Create a Barrier for the indicated number of parties,
   * and the default Rotator function to run at each barrier point.
   */

  this(int parties) { 
    this(parties, &rendezvousFunction); 
  }

  /** 
   * Create a Barrier for the indicated number of parties.
   * and the given function to run at each barrier point.
   */
  this(int parties, RendezvousFunction function_) 
  in { 
    assert(parties > 0);
  }
  body {
    parties_ = parties; 
    rendezvousFunction_ = function_;
    entryGate_ = new FairSemaphore(parties);
    slots_ = new Object[parties];
    wnlock.init();
  }

  /** 
   * Destroy object and release system resources.
   */
  ~this() { 
    wnlock.destroy();
  }

  /**
   * Set the function to call at the point at which all threads reach the
   * rendezvous. This function is run exactly once, by the thread
   * that trips the barrier. The function is not run if the barrier is
   * broken. 
   * /param function the function to run. If null, no function is run.
   * /return the previous function
   */
  synchronized RendezvousFunction setRendezvousFunction(RendezvousFunction function_) {
    RendezvousFunction old = rendezvousFunction_;
    rendezvousFunction_ = function_;
    return old;
  }

  int parties() { return parties_; }

  synchronized bool broken() { return broken_; }

  /**
   * Reset to initial state. Clears both the broken status
   * and any record of waiting threads, and releases all
   * currently waiting threads with indeterminate return status.
   * This method is intended only for use in recovery actions
   * in which it is somehow known
   * that no thread could possibly be relying on the
   * the synchronization properties of this barrier.
   */
  public void restart() { 
    // This is not very good, but probably the best that can be done
    for (;;) {
      synchronized(this) {
        if (entries_ != 0) {
          wnlock.notifyAll();
        }
        else {
          broken_ = false; 
          return;
        }
      }
      Thread.yield();
    }
  }


  /**
   * Enter a rendezvous; returning after all other parties arrive.
   * /param x the item to present at rendezvous point. 
   * By default, this item is exchanged with another.
   * /return an item x given by some thread, and/or processed
   * by the rendezvousFunction.
   * /exception BrokenBarrierException 
   * if any other thread
   * in any previous or current barrier 
   * since either creation or the last <code>restart</code>
   * operation left the barrier
   * prematurely due to interruption or time-out. (If so,
   * the <code>broken</code> status is also set.) 
   * Also returns as
   * broken if the RendezvousFunction encountered a run-time exception.
   */
  Object rendezvous(Object x) {
    return doRendezvous(x, false, 0);
  }

  /**
   * Wait msecs to complete a rendezvous.
   * \param x the item to present at rendezvous point. 
   * By default, this item is exchanged with another.
   * \param msecs The maximum time to wait.
   * \return an item x given by some thread, and/or processed
   * by the rendezvousFunction.
   * \exception BrokenBarrierException 
   * if any other thread
   * in any previous or current barrier 
   * since either creation or the last <code>restart</code>
   * operation left the barrier
   * prematurely due to interruption or time-out. (If so,
   * the <code>broken</code> status is also set.) 
   * Also returns as
   * broken if the RendezvousFunction encountered a run-time exception.
   * \exception WaitException if this thread was interrupted
   * during the exchange. If so, <code>broken</code> status is also set.
   * \exception TimeoutException if this thread timed out waiting for
   * the exchange. If the timeout occured while already in the
   * exchange, <code>broken</code> status is also set.
   */


  Object attemptRendezvous(Object x, long msecs) {
    return doRendezvous(x, true, msecs);
  }

  protected Object doRendezvous(Object x, bool timed, long msecs) {

    // rely on semaphore to throw interrupt on entry

    long startTime;

    if (timed) {
      startTime = clock();
      if (!entryGate_.attempt(msecs)) {
        throw new TimeoutException(msecs);
      }
    }
    else {
      startTime = 0;
      entryGate_.acquire();
    }

    synchronized(this) {

      Object y = null;

      int index =  entries_++;
      slots_[index] = x;

      try { 
        // last one in runs function and releases
        if (entries_ == parties_) {

          departures_ = entries_;
          wnlock.notifyAll();

          try {
            if (!broken_ && rendezvousFunction_ != null)
            rendezvousFunction_(slots_);
          }
          catch (Exception ex) {
            broken_ = true;
          }

        }

        else {

          while (!broken_ && departures_ < 1) {
            long timeLeft = 0;
            if (timed) {
              timeLeft = msecs - (clock() - startTime);
              if (timeLeft <= 0) {
                broken_ = true;
                departures_ = entries_;
                wnlock.notifyAll();
                throw new TimeoutException(msecs);
              }
            }
            
            try {
              wnlock.wait(this,timeLeft); 
            }
            catch (WaitException ex) { 
              if (broken_ || departures_ > 0) {
                break;
              }
              else {
                broken_ = true;
                departures_ = entries_;
                wnlock.notifyAll();
                throw ex;
              }
            }
          }
        }

      }

      finally {

        y = slots_[index];
        
        // Last one out cleans up and allows next set of threads in
        if (--departures_ <= 0) {
          for (int i = 0; i < slots_.length; ++i) slots_[i] = null;
          entryGate_.release(entries_);
          entries_ = 0;
        }
      }

      // continue if no IE/TO throw
      if (broken_)
        throw new BrokenBarrierException(index);
      else
        return y;
    }
  }

}


