/** \file semaphore.d
 * \brief Counting semaphore
 */

 /*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
 */

module concurrent.semaphore;

import concurrent.sync;

private import concurrent.waitnotify;
private import std.c.time;

private import std.thread; // for unittest

/** \class Semaphore
 * \brief Base class for counting semaphores.
 *
 * Conceptually, a semaphore maintains a set of permits.
 * Each acquire() blocks if necessary
 * until a permit is available, and then takes it. 
 * Each release adds a permit. However, no actual permit objects
 * are used; the Semaphore just keeps a count of the number
 * available and acts accordingly.
 * 
 * A semaphore initialized to 1 can serve as a mutual exclusion
 * lock. 
 * 
 * Different implementation subclasses may provide different
 * ordering guarantees (or lack thereof) surrounding which
 * threads will be resumed upon a signal.
 * 
 * The default implementation makes NO 
 * guarantees about the order in which threads will 
 * acquire permits. It is often faster than other implementations.
 * 
 * <b>Sample usage.</b> Here is a class that uses a semaphore to
 * help manage access to a pool of items.
 * \code
 * class Pool {
 *   static final MAX_AVAILABLE = 100;
 *   private final Semaphore available = new Semaphore(MAX_AVAILABLE);
 *   
 *   public Object getItem() throws WaitException { // no synch
 *     available.acquire();
 *     return getNextAvailableItem();
 *   }
 *
 *   public void putItem(Object x) { // no synch
 *     if (markAsUnused(x))
 *       available.release();
 *   }
 *
 *   // Not a particularly efficient data structure; just for demo
 *
 *   protected Object[] items = ... whatever kinds of items being managed
 *   protected bool[] used = new boolean[MAX_AVAILABLE];
 *
 *   protected synchronized Object getNextAvailableItem() { 
 *     for (int i = 0; i < MAX_AVAILABLE; ++i) {
 *       if (!used[i]) {
 *          used[i] = true;
 *          return items[i];
 *       }
 *     }
 *     return null; // not reached 
 *   }
 *
 *   protected synchronized bool markAsUnused(Object item) { 
 *     for (int i = 0; i < MAX_AVAILABLE; ++i) {
 *       if (item == items[i]) {
 *          if (used[i]) {
 *            used[i] = false;
 *            return true;
 *          }
 *          else
 *            return false;
 *       }
 *     }
 *   }
 *   return false;
 *
 * }
 * \endcode
 * 
 */
class Semaphore : Sync  {
  protected long permits_; ///< current number of available permits
  protected SimpleWaitNotify wnlock;

  /** 
   * Create a Semaphore with the given initial number of permits.
   * Using a seed of one makes the semaphore act as a mutual exclusion wnlock.
   * Negative seeds are also allowed, in which case no acquires will proceed
   * until the number of releases has pushed the number of permits past 0.
  */
  this(long initialPermits) {
    permits_ = initialPermits;
    wnlock.init();
  }

  /** Delete semaphore and release system resources.  */
  ~this() {
    wnlock.destroy();
  }

  /** Wait until a permit is available, and take one */
  synchronized void acquire() {
    try {
      while (permits_ <= 0) wnlock.wait(this);
      --permits_;
    }
    catch (WaitException ex) {
      wnlock.notify();
      throw ex;
    }
  }

  /** Wait at most msecs millisconds for a permit. */
  synchronized bool attempt(long msecs) {
    if (permits_ > 0) { 
      --permits_;
      return true;
    }
    else if (msecs <= 0)   
      return false;
    else {
      try {
	long startTime = clock();
	long waitTime = msecs;
          
	for (;;) {
	  wnlock.wait(this,waitTime);
	  if (permits_ > 0) {
	    --permits_;
	    return true;
	  }
	  else { 
	    waitTime = msecs - (clock() - startTime);
	    if (waitTime <= 0) 
	      return false;
	  }
	}
      }
      catch(WaitException ex) { 
	wnlock.notify();
	throw ex;
      }
    }
  }

  /** Release a permit */
  synchronized void release() {
    ++permits_;
    wnlock.notify();
  }


  /** 
   * Release N permits. <code>release(n)</code> is
   * equivalent in effect to:
   * \code
   *   for (int i = 0; i < n; ++i) release();
   * \endcode
   * <p>
   * But may be more efficient in some semaphore implementations.
   */
  synchronized void release(long n) 
  in {
    assert( n >= 0 );
  }
  body {
    permits_ += n;
    for (long i = 0; i < n; ++i) wnlock.notify();
  }

  /**
   * Return the current number of available permits.
   * Returns an accurate, but possibly unstable value,
   * that may change immediately after returning.
   */
  synchronized long permits() {
    return permits_;
  }

  unittest {
    Semaphore sem = new Semaphore(2);
    int done = 0;
    Thread[] t = new Thread[6];
    int f() {
      int n;
      Thread tt = Thread.getThis();
      for (n=0; n < t.length; n++) {
	if (tt is t[n])
	  break;
      }
      printf(" thread %d started\n",n);
      sem.acquire();
      printf(" thread %d aquired\n",n);
      return 0;
    }
    int f2() {
      int n;
      Thread tt = Thread.getThis();
      for (n=0; n < t.length; n++) {
	if (tt is t[n])
	  break;
      }
      printf(" thread %d releasing\n",n);
      sem.release();
      return 0;
    }
    int n;
    for (n=0; n<t.length/2; n++) {
      t[n] = new Thread(&f);
    }
    for (; n<t.length; n++) {
      t[n] = new Thread(&f2);
    }
    printf("starting semaphore unittest\n");
    for (n=0; n<t.length/2; n++) {
      t[n].start();
    }
    Thread.yield();
    for (; n<t.length; n++) {
      t[n].start();
      Thread.yield();
    }
    printf("finished semaphore unittest\n");
    delete sem;
    t[] = null;
  }
}

