/* \file sync
 * \brief Interface for locks, gates and conditions.
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.
  Translated to D by Ben Hinkle 2004.
*/

module concurrent.sync;

/** \class Sync
 * \brief Main interface for locks, gates, and conditions.
 *
 * Sync objects isolate waiting and notification for particular
 * logical states, resource availability, events, and the like that are
 * shared across multiple threads. Use of Syncs sometimes
 * (but by no means always) adds flexibility and efficiency
 * compared to the use of plain D monitor methods
 * and locking, and are sometimes (but by no means always)
 * simpler to program with.
 *
 * Most Syncs are intended to be used primarily (although
 * not exclusively) in  before/after constructions such as:
 * \code
 * class X {
 *   Sync gate;
 *   // ...
 *
 *   public void m() { 
 *     try {
 *       gate.acquire();  // block until condition holds
 *       try {
 *         // ... method body
 *       }
 *       finally {
 *         gate.release()
 *       }
 *     }
 *     catch (WaitException ex) {
 *       // ... evasive action
 *     }
 *   }
 *
 *   public void m2(Sync cond) { // use supplied condition
 *     try {
 *       if (cond.attempt(10)) {         // try the condition for 10 ms
 *         try {
 *           // ... method body
 *         }
 *         finally {
 *           cond.release()
 *         }
 *       }
 *     }
 *     catch (WaitException ex) {
 *       // ... evasive action
 *     }
 *   }
 * }
 * \endcode
 * Syncs may be used in somewhat tedious but more flexible replacements
 * for built-in synchronized blocks. For example:
 * \code
 * class HandSynched {          
 *   private double state_ = 0.0; 
 *   private final Sync lock;  // use lock type supplied in constructor
 *   public HandSynched(Sync l) { lock = l; } 
 *
 *   public void changeState(double d) {
 *     try {
 *       lock.acquire(); 
 *       try     { state_ = updateFunction(d); } 
 *       finally { lock.release(); }
 *     } 
 *     catch(WaitException ex) { }
 *   }
 *
 *   public double getState() {
 *     double d = 0.0;
 *     try {
 *       lock.acquire(); 
 *       try     { d = accessFunction(state_); }
 *       finally { lock.release(); }
 *     } 
 *     catch(WaitException ex){}
 *     return d;
 *   }
 *   private double updateFunction(double d) { ... }
 *   private double accessFunction(double d) { ... }
 * }
 * \endcode
 * 
 * One reason to bother with such constructions is to use deadlock-
 * avoiding back-offs when dealing with locks involving multiple objects.
 * For example, here is a Cell class that uses attempt to back-off
 * and retry if two Cells are trying to swap values with each other 
 * at the same time.
 * \code
 * class Cell {
 *   long value;
 *   Sync lock = ... // some sync implementation class
 *   void swapValue(Cell other) {
 *     for (;;) { 
 *       try {
 *         lock.acquire();
 *         try {
 *           if (other.lock.attempt(100)) {
 *             try { 
 *               long t = value; 
 *               value = other.value;
 *               other.value = t;
 *               return;
 *             }
 *             finally { other.lock.release(); }
 *           }
 *         }
 *         finally { lock.release(); }
 *       } 
 *       catch (WaitException ex) { return; }
 *     }
 *   }
 * }
 * \endcode
 * 
 * Timed versions of attempt report failure via return value.
 * If so desired, you can transform such constructions to use exception
 * throws via 
 * \code
 *   if (!c.attempt(timeval)) throw new TimeoutException(timeval);
 * \endcode
 * 
 * The TimoutSync wrapper class can be used to automate such usages.
 * 
 * All time values are expressed in milliseconds as longs, which have a maximum
 * value of Long.MAX_VALUE, or almost 300,000 centuries.
 * For convenience, some useful time values are defined as static constants.
 * 
 * Syncs may also be used in spinlock constructions. Although
 * it is normally best to just use acquire(), various forms
 * of busy waits can be implemented. For a simple example 
 * (but one that would probably never be preferable to using acquire()):
 * \code
 * class X {
 *   Sync lock = ...
 *   void spinUntilAcquired() {
 *     // Two phase. 
 *     // First spin without pausing.
 *     int purespins = 10; 
 *     for (int i = 0; i < purespins; ++i) {
 *       if (lock.attempt(0))
 *         return true;
 *     }
 *     // Second phase - use timed waits
 *     long waitTime = 1; // 1 millisecond
 *     for (;;) {
 *       if (lock.attempt(waitTime))
 *         return true;
 *       else 
 *         waitTime = waitTime * 3 / 2 + 1; // increase 50% 
 *     }
 *   }
 * }
 * \endcode
 * 
 * In addition pure synchronization control, Syncs
 * may be useful in any context requiring before/after methods.
 * For example, you can use an ObservableSync
 * (perhaps as part of a LayeredSync) in order to obtain callbacks
 * before and after each method invocation for a given class.
 * 
 */


public interface Sync {

  /** 
   *  Wait (possibly forever) until successful passage.
   *  Fail only upon interuption. Interruptions always result in
   *  `clean' failures. On failure,  you can be sure that it has not 
   *  been acquired, and that no 
   *  corresponding release should be performed. Conversely,
   *  a normal return guarantees that the acquire was successful.
   */

  public void acquire();

  /** 
   * Wait at most msecs to pass; report whether passed.
   * <p>
   * The method has best-effort semantics:
   * The msecs bound cannot
   * be guaranteed to be a precise upper bound on wait time in D.
   * Implementations generally can only attempt to return as soon as possible
   * after the specified bound. So, msecs arguments should be used in
   * a coarse-grained manner. Further,
   * implementations cannot always guarantee that this method
   * will return at all without blocking indefinitely when used in
   * unintended ways. For example, deadlocks may be encountered
   * when called in an unintended context.
   * <p>
   * \param msecs the number of milleseconds to wait.
   * An argument less than or equal to zero means not to wait at all. 
   * However, this may still require
   * access to a synchronization lock, which can impose unbounded
   * delay if there is a lot of contention among threads.
   * \return true if acquired
   */

  public bool attempt(long msecs);

  /** 
   * Potentially enable others to pass.
   * <p>
   * Because release does not raise exceptions, 
   * it can be used in `finally' clauses without requiring extra
   * embedded try/catch blocks. But keep in mind that
   * as with any method, implementations may 
   * still throw unchecked exceptions such as Error or NullPointerException
   * when faced with uncontinuable errors. However, these should normally
   * only be caught by higher-level error handlers.
   */

  public void release();

  /**  One second, in milliseconds; convenient as a time-out value */
  public const long ONE_SECOND = 1000;

  /**  One minute, in milliseconds; convenient as a time-out value */
  public const long ONE_MINUTE = 60 * ONE_SECOND;

  /**  One hour, in milliseconds; convenient as a time-out value */
  public const long ONE_HOUR = 60 * ONE_MINUTE;

  /**  One day, in milliseconds; convenient as a time-out value */
  public const long ONE_DAY = 24 * ONE_HOUR;

  /**  One week, in milliseconds; convenient as a time-out value */
  public const long ONE_WEEK = 7 * ONE_DAY;

  /**  One year in milliseconds; convenient as a time-out value  */
  public const long ONE_YEAR = cast(long)(365.2425 * ONE_DAY);

  /**  One century in milliseconds; convenient as a time-out value */
  public const long ONE_CENTURY = 100 * ONE_YEAR;


}

/** \class WaitException
 *
 *  Exception class thrown when a wait() method call fails.
 */
class WaitException : Exception {
  public this(char[] str) { super(str); }
}

