/* \file puttable
 * \brief Interface for input to inter-thread queues
 */

/*
  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.

  History:
  Date       Who                What
  11Jun1998  dl               Create public version
	07May2004  Mike Swieton     Translated to D
*/

module concurrent.takable;

/** 
 * This interface exists to enable stricter type checking for channels. A
 * method argument or instance variable in a consumer object can be declared as
 * only a Takable rather than a Channel, in which case a compiler will disallow
 * put operations.
 * <p>
 * Full method descriptions appear in the Channel interface.
 * @see Channel
 * @see Puttable
**/

interface Takable(T) {

  /** 
   * Return and remove an item from channel, 
   * possibly waiting indefinitely until
   * such an item exists.
   * @return  some item from the channel. Different implementations
   *  may guarantee various properties (such as FIFO) about that item
   *
  **/
  public T take();


  /** 
   * Return and remove an item from channel only if one is available within
   * msecs milliseconds. The time bound is interpreted in a coarse
   * grained, best-effort fashion.
   * @param msecs the number of milliseconds to wait. If less than
   *  or equal to zero, the operation does not perform any timed waits,
   * but might still require
   * access to a synchronization lock, which can impose unbounded
   * delay if there is a lot of contention for the channel.
   * @return some item, or null if the channel is empty.
  **/

  public T poll(long msecs);

}
