
/** \file waitnotify.d
 *  \brief Support wait/notify synchronization and wait/notify/notifyAll
 *
 *  This module exports a struct SimpleWaitNotify for wait() and
 *  notify() and a class SimpleWaitNotifyAll that implements
 *  wait() and notifyAll(). 
 *
 *  In general the user model is similar to Java's
 *  wait/notify/notifyAll user model except that the synchronized
 *  object must be passed to the wait method and the methods are on
 *  the lock object.
 */

/*  Originally written by Ben Hinkle for use in the D port of Doug Lea's
 *  concurrent Java package and released into the public domain.
 *  This may be used for any purposes whatsoever without acknowledgment.
 */

module concurrent.waitnotify;

private import std.thread;
private import concurrent.waitnotifyimpl;
private import concurrent.sync;

/** \class SimpleWaitNotify
 * \brief Struct for supporting wait() and notify() synchronization.
 * 
 *  SimpleWaitNotify adds support for Java-style wait() and notify()
 *  methods. It is similar also to POSIX threading functions cond_wait
 *  and cond_signal.
 *
 *  An example of a class using SimpleWaitNotify:
 *  \code
 *    class A {
 *      SimpleWaitNotify lock;
 *      bool done;
 *      this()  { lock.init(); }
 *      ~this() { lock.destroy(); }
 *      void doSomething1() {
 *        synchronized(this) {
 *          ...
 *          while (!done)
 *            lock.wait(this);
 *          ...
 *        }
 *      }
 *      void doSomething2() {
 *        synchronized(this) {
 *          ...
 *          done = true;
 *          lock.notify();
 *          ...
 *        }
 *      }
 *    }
 *  \endcode
 */
struct SimpleWaitNotify {
  SimpleWaitNotifyImpl impl;

  /** Initialize SimpleWaitNotify
   */
  void init() {
    impl.init();
  }

  /** Destroy SimpleWaitNotify
   */
  void destroy() {
    impl.destroy();
  }

  /** Causes current thread to wait until another thread invokes the notify() method.
   *
   * Causes current thread to wait until another thread invokes the
   * notify() method method for this lock.
   *
   * The current thread must own the supplied object's monitor. The
   * thread releases ownership of the monitor and waits until another
   * thread notifies threads waiting on this lock to wake up through a
   * call to the notify method. The thread then waits until it can
   * re-obtain ownership of the monitor and resumes execution.
   *
   * This method should only be called by a thread that is the owner
   * of the object's monitor. See the notify method for a description
   * of the ways in which a thread can become the owner of a monitor.
   *
   * \param Object obj the shared object to synchronize on
   */
  void wait(Object obj) {
    impl.wait(obj);
  }


  /** Causes current thread to wait until another thread invokes the notify() method.
   *
   * Causes current thread to wait until either another thread invokes
   * the notify() method method for this lock, or a specified amount
   * of time has elapsed.
   *
   * The current thread must own the supplied object's monitor. 
   *
   * This method should only be called by a thread that is the owner
   * of the shared object's monitor. See the notify method for a description
   * of the ways in which a thread can become the owner of a monitor.
   *
   * \param Object obj the shared object to synchronize on
   * \param uint timeout maximum time to wait in milliseconds
   */
  void wait(Object obj, uint timeout) {
    impl.wait(obj,timeout);
  }

  /** Wakes up a single thread that is waiting on this object.
   *
   *  If any threads are waiting on this lock, one of
   *  them is chosen to be awakened. The choice is arbitrary and
   *  occurs at the discretion of the implementation. A thread waits
   *  on this lock by calling one of the wait methods.
   *
   *  The awakened thread will not be able to proceed until the
   *  current thread relinquishes the lock on the shared object. The
   *  awakened thread will compete in the usual manner with any
   *  other threads that might be actively competing to synchronize
   *  on the object; for example, the awakened thread enjoys no
   *  reliable privilege or disadvantage in being the next thread to
   *  lock the object.
   *
   *  This method should only be called by a thread that is the
   *  owner of the shared object's monitor.
   */
  void notify() {
    impl.notify();
  }

}

/** \class SimpleWaitNotifyAll
 * \brief Struct for supporting wait() and notifyAll() synchronization.
 * 
 *  SimpleWaitNotifyAll adds support for Java-style wait() and notifyAll()
 *  methods. It is similar also to POSIX threading functions cond_wait
 *  and cond_broadcast.
 */
struct SimpleWaitNotifyAll {
  SimpleWaitNotifyAllImpl impl;

  /** Initialize SimpleWaitNotifyAll
   */
  void init() { 
    impl.init(); 
  }

  /** Destroy SimpleWaitNotifyAll
   */
  void destroy() { 
    impl.destroy(); 
  }

  /**  Causes current thread to wait until another thread invokes the notifyAll() method.
   *
   *  Same semantics as SimpleWaitNotify.wait except that notifyAll
   *  will wake up all waiting threads.
   *
   * \param Object obj the shared object to synchronize on
   */
  void wait(Object obj) {
    impl.wait(obj);
  }

  /**  Causes current thread to wait until another thread invokes the notifyAll() method.
   *
   *  Same semantics as SimpleWaitNotify.wait except that notifyAll
   *  will wake up all waiting threads.
   *
   * \param Object obj the shared object to synchronize on
   * \param uint timeout maximum time to wait in milliseconds
   */
  void wait(Object obj, uint timeout) {
    impl.wait(obj,timeout);
  }

  /** Wakes up a single thread that is waiting on this object.
   *
   *  If any threads are waiting on this lock, one of
   *  them is chosen to be awakened. The choice is arbitrary and
   *  occurs at the discretion of the implementation. A thread waits
   *  on this lock by calling one of the wait methods.
   *
   *  The awakened thread will not be able to proceed until the
   *  current thread relinquishes the lock on the shared object. The
   *  awakened thread will compete in the usual manner with any
   *  other threads that might be actively competing to synchronize
   *  on the object; for example, the awakened thread enjoys no
   *  reliable privilege or disadvantage in being the next thread to
   *  lock the object.
   *
   *  This method should only be called by a thread that is the
   *  owner of the shared object's monitor.
   */
  void notify() {
    impl.notify();
  }

  /** Wakes up all threads that are waiting on this object.
   *
   *  Same semantics as SimpleWaitNotify.notify except that notifyAll
   *  will wake up all threads waiting on this lock. The threads will
   *  compete with other threads for the object monitor once this
   *  thread releases the monitor.
   */
  void notifyAll() {
    impl.notifyAll();
  }
}

unittest {

  class A {
    SimpleWaitNotify lock;
    bool done, finished1, finished2;
    this()  { lock.init(); }
    ~this() { lock.destroy(); }
    synchronized int doSomething1() {
      while (!done) {
	lock.wait(this);
      }
      finished1 = 1;
      return 0;
    }
    synchronized int doSomething2() {
      done = true;
      lock.notify();
      finished2 = 1;
      return 0;
    }
  } // A
  printf("starting waitnotify unittest\n");
  A a = new A();
  Thread t1 = new Thread(&a.doSomething1);
  Thread t2 = new Thread(&a.doSomething2);
  t1.start();
  t2.start();
  while (!a.finished1 && !a.finished2)
    Thread.yield();
  //  delete a;  // causes errors with DMD on linux
  printf("finished waitnotify unittest\n");
  assert(true);
}

