
/** \file waitnotifyimpl.d
 *  \brief Platform and compiler-specific implementation of WaitNotify
 *
 *  This module supplies the implementation for different platforms 
 *  and compilers for concurrent.waitnotify. The internal Object layout
 *  and the Object monitor implementation are used.
 */

/*  Originally written by Ben Hinkle for use in the D port of Doug Lea's
 *  concurrent Java package and released into the public domain.
 *  This may be used for any purposes whatsoever without acknowledgment.
 */

module concurrent.waitnotifyimpl;

private import std.thread;
private import concurrent.sync;

version (Windows) {
  private import std.c.windows.windows;

  version = ATOMIC; 

  private void waitWin32Impl(HANDLE event, Object obj, uint timeout) {
    // how can we check that thread has obj's monitor?
    if (timeout == 0) 
      timeout = INFINITE; 
    InternalObjectRep* iobj = cast(InternalObjectRep*)obj;
    LONG RecursionCount = iobj.monitor.RecursionCount;
    DWORD firedEvent;
    version (ATOMIC) {
      /* Atomically release the LockSemaphore (if present) and wait.
       * Since the SpinCount is always 0 for D critical sections the
       * LockSemaphore will always be around if someone is waiting
       * for this critical section.
       */
      iobj.monitor.RecursionCount = 0;
      iobj.monitor.OwningThread = cast(HANDLE)0;
      for (int k=1;k<RecursionCount;k++) {
	InterlockedDecrement(&iobj.monitor.LockCount);
      }
      // hmm. this probably still isn't atomic since 
      // EnterCriticalSection most likely checks the LockCount
      // before waiting on the LockSemaphore.
      if (InterlockedDecrement(&iobj.monitor.LockCount) >= 0) {
	firedEvent = SignalObjectAndWait(iobj.monitor.LockSemaphore,
					 event,timeout,0);
      } else {
	firedEvent = WaitForSingleObject(event,timeout);
      }
    } else {
      iobj.monitor.RecursionCount = 1;
      for (int k=1;k<RecursionCount;k++) {
	InterlockedDecrement(&iobj.monitor.LockCount);
      }
      _d_monitorexit(obj);
      firedEvent = WaitForSingleObject(event,timeout);
    }
    _d_monitorenter(obj);
    iobj.monitor.RecursionCount = RecursionCount;
    for (int k=1;k<RecursionCount;k++) {
      InterlockedIncrement(&iobj.monitor.LockCount);
    }
    if (firedEvent == WAIT_FAILED) {
      throw new WaitException("Wait failed");
    }
  }

  struct SimpleWaitNotifyImpl {
    private HANDLE event;

    void init() {
      event = CreateEventA(null,0,0,null);
    }

    void destroy() {
      CloseHandle(event);
    }

    void wait(Object obj) {
      wait(obj,0);
    }

    void wait(Object obj, uint timeout) {
      waitWin32Impl(event,obj,timeout);
    }

    void notify() {
      SetEvent(event); // wakes one thread and resets event
    }
  }

  struct SimpleWaitNotifyAllImpl {
    private HANDLE event;

    void init() {
      event = CreateEventA(null,1,0,null);
    }

    void destroy() {
      CloseHandle(event);
    }

    void wait(Object obj) {
      wait(obj,0);
    }

    void wait(Object obj, uint timeout) {
      waitWin32Impl(event,obj,timeout);
    }

    void notify() {
      SetEvent(event); // wakes one thread and resets event
    }

    void notifyAll() {
      PulseEvent(event); // wakes all thread and resets event
    }
  }

  private {
    extern (Windows) {
      HANDLE CreateEventA(LPSECURITY_ATTRIBUTES, BOOL, BOOL, LPCSTR);
      DWORD WaitForSingleObject(HANDLE, DWORD);
      DWORD SignalObjectAndWait(HANDLE,HANDLE, DWORD,BOOL);
      DWORD SetEvent(HANDLE);
      DWORD PulseEvent(HANDLE);
      DWORD ResetEvent(HANDLE);
      DWORD WaitForMultipleObjects(DWORD, HANDLE*, BOOL, DWORD);
    }
    const DWORD INFINITE = -1;
    const DWORD WAIT_FAILED = -1;
    struct RTL_CRITICAL_SECTION {
      void* DebugInfo;
      LONG LockCount;
      LONG RecursionCount;
      HANDLE OwningThread;
      HANDLE LockSemaphore;
      // ignore the rest
    }
    /* D Object layout puts monitor after vtbl */
    struct InternalObjectRep {
      void* vtbl;
      RTL_CRITICAL_SECTION* monitor;
    }
  }

} else version (linux) {

  private {
    import std.c.linux.linux;
    import std.c.linux.linuxextern;
		import std.string;
  }

  struct SimpleWaitNotifyImpl {
    private ubyte[48] cond;    // _pthread_cond_t has sizeof 48

    void init() {
			//memset(cond, 0, 48);
      pthread_cond_init(cond,null);
    }
    void destroy() {
      pthread_cond_destroy(cond);
    }

    void wait(Object obj) {
      // how can we check that thread has obj's monitor?
      InternalObjectRep* iobj = cast(InternalObjectRep*)obj;
      int RecursionCount = iobj.monitor.__m_count;
      iobj.monitor.__m_count = 1;
      int res = pthread_cond_wait(cond,iobj.monitor);
      iobj.monitor.__m_count = RecursionCount;
      if (res) {
      	throw new WaitException("Wait failed");
      }
    }

    void wait(Object obj, uint timeout) {
      // how can we check that thread has obj's monitor?
      InternalObjectRep* iobj = cast(InternalObjectRep*)obj;
      timespec abstime;
      timeval tv;
      gettimeofday(&tv,null);
      long micro = tv.tv_usec+timeout*1000;
      abstime.tv_sec = tv.tv_sec+micro/1000000;
      abstime.tv_nsec = (micro%1000000)*1000;
      int RecursionCount = iobj.monitor.__m_count;
      iobj.monitor.__m_count = 1;
      int res = pthread_cond_timedwait(cond,iobj.monitor,&abstime);
      if (res) {
	_d_monitorenter(obj);
	iobj.monitor.__m_count = RecursionCount;
	if (res != ETIMEDOUT) {
	  throw new WaitException("Wait failed");
	}
      } else {
	iobj.monitor.__m_count = RecursionCount;
      }
    }

    void notify() {
      pthread_cond_signal(cond);
    }

  }

  struct SimpleWaitNotifyAllImpl {
    SimpleWaitNotifyImpl wnlock;
    void init() {
      wnlock.init();
    }
    void destroy() {
      wnlock.destroy();
    }
    void wait(Object obj) {
      wnlock.wait(obj);
    }
    void wait(Object obj, uint timeout) {
      wnlock.wait(obj,timeout);
    }

    void notify() {
			wnlock.notify();
    }

    void notifyAll() {
      pthread_cond_broadcast(wnlock.cond);
    }
  }

  private {
    struct timespec {
      time_t tv_sec;		/* Seconds.  */
      int tv_nsec;		/* Nanoseconds.  */
    }
    /* from bits/pthreadtypes.h */
    struct pthread_mutex_t {
      int __m_reserved;       
      int __m_count;          
      // ignore the rest
    };

    /* D Object layout puts monitor after vtbl */
    struct InternalObjectRep {
      void* vtbl;
      pthread_mutex_t* monitor;
    }
    const int ETIMEDOUT = 110; // from errno.h
    extern (C) {
      int pthread_cond_init(void* cond, void* attr);
      int pthread_cond_destroy(void* cond);
      int pthread_cond_wait(void* cond, void* mutex);
      int pthread_cond_timedwait(void* cond, void* mutex, timespec* abstime);
      int pthread_cond_signal(void* cond);
      int pthread_cond_broadcast(void* cond);
    }
  }
}

/* Internal D functions to aquire/release object monitor */
private {
  extern (C) void _d_monitorexit(void* h);
  extern (C) void _d_monitorenter(void* h);
}

